@@ -3,7 +3,7 @@
 ** Copyright (C) 2015 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing
 **
-** This file is part of the Qt Build Suite.
+** This file is part of Qbs.
 **
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
@@ -44,6 +44,7 @@
 #include <language/language.h>
 #include <language/loader.h>
 #include <language/propertymapinternal.h>
+#include <language/resolvedfilecontext.h>
 #include <logging/translator.h>
 #include <tools/fileinfo.h>
 #include <tools/persistence.h>
@@ -218,6 +219,7 @@ void BuildGraphLoader::trackProjectChanges()
             || hasEnvironmentChanged(restoredProject)
             || hasCanonicalFilePathResultChanged(restoredProject)
             || hasFileExistsResultChanged(restoredProject)
+            || hasDirectoryEntriesResultChanged(restoredProject)
             || hasFileLastModifiedResultChanged(restoredProject)) {
         reResolvingNecessary = true;
     }
@@ -380,6 +382,23 @@ bool BuildGraphLoader::hasFileExistsResultChanged(const TopLevelProjectConstPtr
     return false;
 }
 
+bool BuildGraphLoader::hasDirectoryEntriesResultChanged(const TopLevelProjectConstPtr &restoredProject) const
+{
+    for (auto it = restoredProject->directoryEntriesResults.constBegin();
+         it != restoredProject->directoryEntriesResults.constEnd(); ++it) {
+        if (QDir(it.key().first).entryList(static_cast<QDir::Filters>(it.key().second), QDir::Name)
+                != it.value()) {
+            m_logger.qbsDebug() << "Entry list for directory '" << it.key().first
+                                << "' ("
+                                << static_cast<QDir::Filters>(it.key().second)
+                                << ") changed, must re-resolve project.";
+            return true;
+        }
+    }
+
+    return false;
+}
+
 bool BuildGraphLoader::hasFileLastModifiedResultChanged(const TopLevelProjectConstPtr &restoredProject) const
 {
     for (QHash<QString, FileTime>::ConstIterator it
@@ -497,17 +516,29 @@ bool BuildGraphLoader::checkProductForChanges(const ResolvedProductPtr &restored
                                               const ResolvedProductPtr &newlyResolvedProduct)
 {
     // This check must come first, as it can prevent build data rescuing as a side effect.
-    // TODO: Similar special checks must be done for qbs.getEnv() and File.exists() in
+    // TODO: Similar special checks must be done for Environment.getEnv() and File.exists() in
     // commands (or possibly it could be reasonable to just forbid such "dynamic" constructs
     // within commands).
     if (checkForPropertyChanges(restoredProduct, newlyResolvedProduct))
         return true;
-
-    return !transformerListsAreEqual(restoredProduct->transformers,
-                                     newlyResolvedProduct->transformers)
-            || !ruleListsAreEqual(restoredProduct->rules.toList(),
-                                  newlyResolvedProduct->rules.toList())
-            || !dependenciesAreEqual(restoredProduct, newlyResolvedProduct);
+    if (!transformerListsAreEqual(restoredProduct->transformers,
+                                  newlyResolvedProduct->transformers)) {
+        return true;
+    }
+    if (!ruleListsAreEqual(restoredProduct->rules.toList(), newlyResolvedProduct->rules.toList()))
+        return true;
+    if (!dependenciesAreEqual(restoredProduct, newlyResolvedProduct))
+        return true;
+    const FileTime referenceTime = restoredProduct->topLevelProject()->lastResolveTime;
+    foreach (const ResolvedTransformerConstPtr &transformer, newlyResolvedProduct->transformers) {
+        if (!isPrepareScriptUpToDate(transformer->transform, referenceTime))
+            return true;
+    }
+    foreach (const RuleConstPtr &rule, newlyResolvedProduct->rules) {
+        if (!isPrepareScriptUpToDate(rule->prepareScript, referenceTime))
+            return true;
+    }
+    return false;
 }
 
 bool BuildGraphLoader::checkProductForInstallInfoChanges(const ResolvedProductPtr &restoredProduct,
@@ -738,6 +769,21 @@ bool BuildGraphLoader::isConfigCompatible()
     return true;
 }
 
+bool BuildGraphLoader::isPrepareScriptUpToDate(const ScriptFunctionConstPtr &script,
+                                               const FileTime &referenceTime)
+{
+    foreach (const JsImport &jsImport, script->fileContext->jsImports()) {
+        foreach (const QString &filePath, jsImport.filePaths) {
+            if (FileInfo(filePath).lastModified() > referenceTime) {
+                m_logger.qbsDebug() << "Change in import '" << filePath
+                        << "' potentially influences prepare script, marking as out of date";
+                return false;
+            }
+        }
+    }
+    return true;
+}
+
 void BuildGraphLoader::rescueOldBuildData(const ResolvedProductConstPtr &restoredProduct,
         const ResolvedProductPtr &newlyResolvedProduct, ProjectBuildData *oldBuildData,
         const ChildListHash &childLists, const AllRescuableArtifactData &existingRad)
