@@ -3,7 +3,7 @@
 ** Copyright (C) 2015 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing
 **
-** This file is part of the Qt Build Suite.
+** This file is part of Qbs.
 **
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
@@ -33,9 +33,14 @@
 #include "deprecationinfo.h"
 
 #include <logging/translator.h>
+#include <tools/error.h>
+#include <tools/qbsassert.h>
 
 #include <QStringList>
 
+#include <initializer_list>
+#include <utility>
+
 namespace qbs {
 namespace Internal {
 
@@ -46,6 +51,23 @@ const char QBS_LANGUAGE_VERSION[] = "1.0";
 
 BuiltinDeclarations::BuiltinDeclarations()
     : m_languageVersion(Version::fromString(QLatin1String(QBS_LANGUAGE_VERSION)))
+    , m_typeMap(std::initializer_list<std::pair<QString, ItemType>>({
+        { QLatin1String("Artifact"), ItemType::Artifact },
+        { QLatin1String("Depends"), ItemType::Depends },
+        { QLatin1String("Export"), ItemType::Export },
+        { QLatin1String("FileTagger"), ItemType::FileTagger },
+        { QLatin1String("Group"), ItemType::Group },
+        { QLatin1String("Module"), ItemType::Module },
+        { QLatin1String("Probe"), ItemType::Probe },
+        { QLatin1String("Product"), ItemType::Product },
+        { QLatin1String("Project"), ItemType::Project },
+        { QLatin1String("Properties"), ItemType::Properties }, // Callers have to handle the SubProject case.
+        { QLatin1String("PropertyOptions"), ItemType::PropertyOptions },
+        { QLatin1String("Rule"), ItemType::Rule },
+        { QLatin1String("Scanner"), ItemType::Scanner },
+        { QLatin1String("SubProject"), ItemType::SubProject },
+        { QLatin1String("Transformer"), ItemType::Transformer }
+    }))
 {
     addArtifactItem();
     addDependsItem();
@@ -74,24 +96,41 @@ Version BuiltinDeclarations::languageVersion() const
     return m_languageVersion;
 }
 
-bool BuiltinDeclarations::containsType(const QString &typeName) const
-{
-    return m_builtins.contains(typeName);
-}
-
 QStringList BuiltinDeclarations::allTypeNames() const
 {
-    return m_builtins.keys();
+    return m_typeMap.keys();
 }
 
-ItemDeclaration BuiltinDeclarations::declarationsForType(const QString &typeName) const
+ItemDeclaration BuiltinDeclarations::declarationsForType(ItemType type) const
 {
-    return m_builtins.value(typeName);
+    return m_builtins.value(type);
+}
+
+ItemType BuiltinDeclarations::typeForName(const QString &typeName,
+                                          const CodeLocation location) const
+{
+    const auto it = m_typeMap.constFind(typeName);
+    if (it == m_typeMap.constEnd())
+        throw ErrorInfo(Tr::tr("Unexpected item type '%1'.").arg(typeName), location);
+    return it.value();
+}
+
+QString BuiltinDeclarations::nameForType(ItemType itemType) const
+{
+    // Iterating is okay here, as this mapping is not used in hot code paths.
+    if (itemType == ItemType::PropertiesInSubProject)
+        return QLatin1String("Properties");
+    for (auto it = m_typeMap.constBegin(); it != m_typeMap.constEnd(); ++it) {
+        if (it.value() == itemType)
+            return it.key();
+    }
+    QBS_CHECK(false);
+    return QString();
 }
 
 void BuiltinDeclarations::insert(const ItemDeclaration &decl)
 {
-    m_builtins.insert(decl.typeName(), decl);
+    m_builtins.insert(decl.type(), decl);
 }
 
 static PropertyDeclaration conditionProperty()
@@ -101,6 +140,13 @@ static PropertyDeclaration conditionProperty()
     return decl;
 }
 
+static PropertyDeclaration alwaysRunProperty()
+{
+    PropertyDeclaration decl(QLatin1String("alwaysRun"), PropertyDeclaration::Boolean);
+    decl.setInitialValueSource(QLatin1String("false"));
+    return decl;
+}
+
 static PropertyDeclaration nameProperty()
 {
     return PropertyDeclaration(QLatin1String("name"), PropertyDeclaration::String);
@@ -113,7 +159,8 @@ static PropertyDeclaration buildDirProperty()
 
 static PropertyDeclaration prepareScriptProperty()
 {
-    PropertyDeclaration decl(QLatin1String("prepare"), PropertyDeclaration::Verbatim);
+    PropertyDeclaration decl(QLatin1String("prepare"), PropertyDeclaration::Variant,
+                             PropertyDeclaration::PropertyNotAvailableInConfig);
     decl.setFunctionArgumentNames(
                 QStringList()
                 << QLatin1String("project") << QLatin1String("product")
@@ -124,17 +171,17 @@ static PropertyDeclaration prepareScriptProperty()
 
 void BuiltinDeclarations::addArtifactItem()
 {
-    ItemDeclaration item(QLatin1String("Artifact"));
+    ItemDeclaration item(ItemType::Artifact);
     PropertyDeclaration conditionDecl = conditionProperty();
     conditionDecl.setDeprecationInfo(DeprecationInfo(Version(1, 4), Tr::tr("If you need "
             "dynamic artifacts, use the Rule.outputArtifacts script instead of Artifact items.")));
     item << conditionDecl;
-    PropertyDeclaration fileNameDecl(QLatin1String("fileName"), PropertyDeclaration::Verbatim);
+    PropertyDeclaration fileNameDecl(QLatin1String("fileName"), PropertyDeclaration::String);
     fileNameDecl.setDeprecationInfo(DeprecationInfo(Version(1, 4),
                                                     Tr::tr("Please use 'filePath' instead.")));
     item << fileNameDecl;
-    item << PropertyDeclaration(QLatin1String("filePath"), PropertyDeclaration::Verbatim);
-    item << PropertyDeclaration(QLatin1String("fileTags"), PropertyDeclaration::Variant);
+    item << PropertyDeclaration(QLatin1String("filePath"), PropertyDeclaration::String);
+    item << PropertyDeclaration(QLatin1String("fileTags"), PropertyDeclaration::StringList);
     PropertyDeclaration decl(QLatin1String("alwaysUpdated"), PropertyDeclaration::Boolean);
     decl.setInitialValueSource(QLatin1String("true"));
     item << decl;
@@ -143,10 +190,10 @@ void BuiltinDeclarations::addArtifactItem()
 
 void BuiltinDeclarations::addDependsItem()
 {
-    ItemDeclaration item(QLatin1String("Depends"));
+    ItemDeclaration item(ItemType::Depends);
     item << conditionProperty();
     item << nameProperty();
-    item << PropertyDeclaration(QLatin1String("submodules"), PropertyDeclaration::Variant);
+    item << PropertyDeclaration(QLatin1String("submodules"), PropertyDeclaration::StringList);
     PropertyDeclaration requiredDecl(QLatin1String("required"), PropertyDeclaration::Boolean);
     requiredDecl.setInitialValueSource(QLatin1String("true"));
     item << requiredDecl;
@@ -162,32 +209,32 @@ void BuiltinDeclarations::addDependsItem()
 
 void BuiltinDeclarations::addExportItem()
 {
-    addModuleLikeItem(QStringLiteral("Export"));
+    addModuleLikeItem(ItemType::Export);
 }
 
 void BuiltinDeclarations::addFileTaggerItem()
 {
-    ItemDeclaration item(QLatin1String("FileTagger"));
+    ItemDeclaration item(ItemType::FileTagger);
     item << PropertyDeclaration(QLatin1String("patterns"), PropertyDeclaration::StringList);
-    item << PropertyDeclaration(QLatin1String("fileTags"), PropertyDeclaration::Variant);
+    item << PropertyDeclaration(QLatin1String("fileTags"), PropertyDeclaration::StringList);
     insert(item);
 }
 
 void BuiltinDeclarations::addGroupItem()
 {
-    ItemDeclaration item(QLatin1String("Group"));
+    ItemDeclaration item(ItemType::Group);
     item << conditionProperty();
     item << PropertyDeclaration(QLatin1String("name"), PropertyDeclaration::String,
                                 PropertyDeclaration::PropertyNotAvailableInConfig);
-    item << PropertyDeclaration(QLatin1String("files"), PropertyDeclaration::Variant,
+    item << PropertyDeclaration(QLatin1String("files"), PropertyDeclaration::PathList,
                                 PropertyDeclaration::PropertyNotAvailableInConfig);
-    item << PropertyDeclaration(QLatin1String("fileTagsFilter"), PropertyDeclaration::Variant,
+    item << PropertyDeclaration(QLatin1String("fileTagsFilter"), PropertyDeclaration::StringList,
                                 PropertyDeclaration::PropertyNotAvailableInConfig);
-    item << PropertyDeclaration(QLatin1String("excludeFiles"), PropertyDeclaration::Variant,
+    item << PropertyDeclaration(QLatin1String("excludeFiles"), PropertyDeclaration::PathList,
                                 PropertyDeclaration::PropertyNotAvailableInConfig);
-    item << PropertyDeclaration(QLatin1String("fileTags"), PropertyDeclaration::Variant,
+    item << PropertyDeclaration(QLatin1String("fileTags"), PropertyDeclaration::StringList,
                                 PropertyDeclaration::PropertyNotAvailableInConfig);
-    item << PropertyDeclaration(QLatin1String("prefix"), PropertyDeclaration::Variant,
+    item << PropertyDeclaration(QLatin1String("prefix"), PropertyDeclaration::String,
                                 PropertyDeclaration::PropertyNotAvailableInConfig);
     PropertyDeclaration declaration;
     declaration.setName(QLatin1String("overrideTags"));
@@ -200,32 +247,34 @@ void BuiltinDeclarations::addGroupItem()
 
 void BuiltinDeclarations::addModuleItem()
 {
-    addModuleLikeItem(QStringLiteral("Module"));
+    addModuleLikeItem(ItemType::Module);
 }
 
-void BuiltinDeclarations::addModuleLikeItem(const QString &typeName)
+void BuiltinDeclarations::addModuleLikeItem(ItemType type)
 {
-    ItemDeclaration item(typeName);
+    ItemDeclaration item(type);
     item.setAllowedChildTypes(ItemDeclaration::TypeNames()
-            << QLatin1String("Group")
-            << QLatin1String("Depends")
-            << QLatin1String("FileTagger")
-            << QLatin1String("Rule")
-            << QLatin1String("Probe")
-            << QLatin1String("PropertyOptions")
-            << QLatin1String("Transformer")
-            << QLatin1String("Scanner"));
+            << ItemType::Group
+            << ItemType::Depends
+            << ItemType::FileTagger
+            << ItemType::Rule
+            << ItemType::Probe
+            << ItemType::PropertyOptions
+            << ItemType::Transformer
+            << ItemType::Scanner);
     item << nameProperty();
     item << conditionProperty();
     item << PropertyDeclaration(QLatin1String("setupBuildEnvironment"),
-                                      PropertyDeclaration::Verbatim);
-    item << PropertyDeclaration(QLatin1String("setupRunEnvironment"),
-                                      PropertyDeclaration::Verbatim);
-    item << PropertyDeclaration(QLatin1String("validate"),
                                       PropertyDeclaration::Variant,
                                       PropertyDeclaration::PropertyNotAvailableInConfig);
+    item << PropertyDeclaration(QLatin1String("setupRunEnvironment"),
+                                      PropertyDeclaration::Variant,
+                                      PropertyDeclaration::PropertyNotAvailableInConfig);
+    item << PropertyDeclaration(QLatin1String("validate"),
+                                      PropertyDeclaration::Boolean,
+                                      PropertyDeclaration::PropertyNotAvailableInConfig);
     item << PropertyDeclaration(QLatin1String("additionalProductTypes"),
-                                      PropertyDeclaration::Variant);
+                                      PropertyDeclaration::StringList);
     PropertyDeclaration presentDecl(QLatin1String("present"), PropertyDeclaration::Boolean);
     presentDecl.setInitialValueSource(QLatin1String("true"));
     item << presentDecl;
@@ -234,27 +283,28 @@ void BuiltinDeclarations::addModuleLikeItem(const QString &typeName)
 
 void BuiltinDeclarations::addProbeItem()
 {
-    ItemDeclaration item(QLatin1String("Probe"));
+    ItemDeclaration item(ItemType::Probe);
     item << conditionProperty();
     PropertyDeclaration foundProperty(QLatin1String("found"), PropertyDeclaration::Boolean);
     foundProperty.setInitialValueSource(QLatin1String("false"));
     item << foundProperty;
-    item << PropertyDeclaration(QLatin1String("configure"), PropertyDeclaration::Verbatim);
+    item << PropertyDeclaration(QLatin1String("configure"), PropertyDeclaration::Variant,
+                                PropertyDeclaration::PropertyNotAvailableInConfig);
     insert(item);
 }
 
 void BuiltinDeclarations::addProductItem()
 {
-    ItemDeclaration item(QLatin1String("Product"));
+    ItemDeclaration item(ItemType::Product);
     item.setAllowedChildTypes(ItemDeclaration::TypeNames()
-            << QLatin1String("Depends")
-            << QLatin1String("Transformer")
-            << QLatin1String("Group")
-            << QLatin1String("FileTagger")
-            << QLatin1String("Export")
-            << QLatin1String("Probe")
-            << QLatin1String("PropertyOptions")
-            << QLatin1String("Rule"));
+            << ItemType::Depends
+            << ItemType::Transformer
+            << ItemType::Group
+            << ItemType::FileTagger
+            << ItemType::Export
+            << ItemType::Probe
+            << ItemType::PropertyOptions
+            << ItemType::Rule);
     item << conditionProperty();
     PropertyDeclaration decl(QLatin1String("type"), PropertyDeclaration::StringList);
     decl.setInitialValueSource(QLatin1String("[]"));
@@ -277,9 +327,9 @@ void BuiltinDeclarations::addProductItem()
     item << decl;
     item << PropertyDeclaration(QLatin1String("consoleApplication"),
                                 PropertyDeclaration::Boolean);
-    item << PropertyDeclaration(QLatin1String("files"), PropertyDeclaration::Variant,
+    item << PropertyDeclaration(QLatin1String("files"), PropertyDeclaration::PathList,
                                 PropertyDeclaration::PropertyNotAvailableInConfig);
-    item << PropertyDeclaration(QLatin1String("excludeFiles"), PropertyDeclaration::Variant,
+    item << PropertyDeclaration(QLatin1String("excludeFiles"), PropertyDeclaration::PathList,
                                 PropertyDeclaration::PropertyNotAvailableInConfig);
     item << PropertyDeclaration(QLatin1String("qbsSearchPaths"),
                                 PropertyDeclaration::StringList);
@@ -289,21 +339,21 @@ void BuiltinDeclarations::addProductItem()
 
 void BuiltinDeclarations::addProjectItem()
 {
-    ItemDeclaration item(QLatin1String("Project"));
+    ItemDeclaration item(ItemType::Project);
     item.setAllowedChildTypes(ItemDeclaration::TypeNames()
-            << QLatin1String("Project")
-            << QLatin1String("PropertyOptions")
-            << QLatin1String("SubProject")
-            << QLatin1String("Product")
-            << QLatin1String("FileTagger")
-            << QLatin1String("Rule"));
+            << ItemType::Project
+            << ItemType::PropertyOptions
+            << ItemType::SubProject
+            << ItemType::Product
+            << ItemType::FileTagger
+            << ItemType::Rule);
     item << nameProperty();
     item << conditionProperty();
     item << buildDirProperty();
     item << PropertyDeclaration(QLatin1String("minimumQbsVersion"), PropertyDeclaration::String);
     item << PropertyDeclaration(QLatin1String("sourceDirectory"), PropertyDeclaration::Path);
     item << PropertyDeclaration(QLatin1String("profile"), PropertyDeclaration::String);
-    item << PropertyDeclaration(QLatin1String("references"), PropertyDeclaration::Variant,
+    item << PropertyDeclaration(QLatin1String("references"), PropertyDeclaration::PathList,
                                       PropertyDeclaration::PropertyNotAvailableInConfig);
     item << PropertyDeclaration(QLatin1String("qbsSearchPaths"),
             PropertyDeclaration::StringList, PropertyDeclaration::PropertyNotAvailableInConfig);
@@ -312,12 +362,12 @@ void BuiltinDeclarations::addProjectItem()
 
 void BuiltinDeclarations::addPropertiesItem()
 {
-    insert(ItemDeclaration(QLatin1String("Properties")));
+    insert(ItemDeclaration(ItemType::Properties));
 }
 
 void BuiltinDeclarations::addPropertyOptionsItem()
 {
-    ItemDeclaration item(QLatin1String("PropertyOptions"));
+    ItemDeclaration item(ItemType::PropertyOptions);
     item << nameProperty();
     item << PropertyDeclaration(QLatin1String("allowedValues"), PropertyDeclaration::Variant);
     item << PropertyDeclaration(QLatin1String("description"), PropertyDeclaration::String);
@@ -326,17 +376,19 @@ void BuiltinDeclarations::addPropertyOptionsItem()
 
 void BuiltinDeclarations::addRuleItem()
 {
-    ItemDeclaration item(QLatin1String("Rule"));
+    ItemDeclaration item(ItemType::Rule);
     item.setAllowedChildTypes(ItemDeclaration::TypeNames()
-            << QLatin1String("Artifact"));
+            << ItemType::Artifact);
     item << conditionProperty();
+    item << alwaysRunProperty();
     PropertyDeclaration decl(QLatin1String("multiplex"), PropertyDeclaration::Boolean);
     decl.setInitialValueSource(QLatin1String("false"));
     item << decl;
     item << PropertyDeclaration(QLatin1String("name"), PropertyDeclaration::String);
     item << PropertyDeclaration(QLatin1String("inputs"), PropertyDeclaration::StringList);
     item << PropertyDeclaration(QLatin1String("outputFileTags"), PropertyDeclaration::StringList);
-    decl = PropertyDeclaration(QLatin1String("outputArtifacts"), PropertyDeclaration::Verbatim);
+    decl = PropertyDeclaration(QLatin1String("outputArtifacts"), PropertyDeclaration::Variant,
+                               PropertyDeclaration::PropertyNotAvailableInConfig);
     decl.setFunctionArgumentNames(
                 QStringList()
                 << QLatin1String("project") << QLatin1String("product")
@@ -359,11 +411,11 @@ void BuiltinDeclarations::addRuleItem()
 
 void BuiltinDeclarations::addSubprojectItem()
 {
-    ItemDeclaration item(QLatin1String("SubProject"));
+    ItemDeclaration item(ItemType::SubProject);
     item.setAllowedChildTypes(ItemDeclaration::TypeNames()
-            << QLatin1String("Project") // needed, because we're adding this internally
-            << QLatin1String("Properties")
-            << QLatin1String("PropertyOptions"));
+            << ItemType::Project // needed, because we're adding this internally
+            << ItemType::PropertiesInSubProject
+            << ItemType::PropertyOptions);
     item << PropertyDeclaration(QLatin1String("filePath"), PropertyDeclaration::Path);
     PropertyDeclaration inheritProperty;
     inheritProperty.setName(QLatin1String("inheritProperties"));
@@ -375,11 +427,12 @@ void BuiltinDeclarations::addSubprojectItem()
 
 void BuiltinDeclarations::addTransformerItem()
 {
-    ItemDeclaration item(QLatin1String("Transformer"));
+    ItemDeclaration item(ItemType::Transformer);
     item.setAllowedChildTypes(ItemDeclaration::TypeNames()
-            << QLatin1String("Artifact"));
+            << ItemType::Artifact);
     item << conditionProperty();
-    item << PropertyDeclaration(QLatin1String("inputs"), PropertyDeclaration::Variant);
+    item << alwaysRunProperty();
+    item << PropertyDeclaration(QLatin1String("inputs"), PropertyDeclaration::PathList);
     item << prepareScriptProperty();
     item << PropertyDeclaration(QLatin1String("explicitlyDependsOn"),
                                       PropertyDeclaration::StringList);
@@ -388,20 +441,21 @@ void BuiltinDeclarations::addTransformerItem()
 
 void BuiltinDeclarations::addScannerItem()
 {
-    ItemDeclaration item(QLatin1String("Scanner"));
+    ItemDeclaration item(ItemType::Scanner);
     item << conditionProperty();
     item << PropertyDeclaration(QLatin1String("inputs"), PropertyDeclaration::StringList);
     PropertyDeclaration recursive(QLatin1String("recursive"), PropertyDeclaration::Boolean);
     recursive.setInitialValueSource(QLatin1String("false"));
     item << recursive;
-    PropertyDeclaration searchPaths(QLatin1String("searchPaths"), PropertyDeclaration::Verbatim);
+    PropertyDeclaration searchPaths(QLatin1String("searchPaths"), PropertyDeclaration::StringList);
     searchPaths.setFunctionArgumentNames(
                 QStringList()
                 << QLatin1String("project")
                 << QLatin1String("product")
                 << QLatin1String("input"));
     item << searchPaths;
-    PropertyDeclaration scan(QLatin1String("scan"), PropertyDeclaration::Verbatim);
+    PropertyDeclaration scan(QLatin1String("scan"), PropertyDeclaration::Variant,
+                             PropertyDeclaration::PropertyNotAvailableInConfig);
     scan.setFunctionArgumentNames(
                 QStringList()
                 << QLatin1String("project")
